"""
Model of files that hold commands.
"""

from __future__ import annotations
import copy
from dataclasses import dataclass, field, InitVar
from pathlib import Path
from textwrap import dedent
from typing import Protocol, cast, overload, TYPE_CHECKING
from typing_extensions import Final, override

from hpcflow.sdk.core.types import ActionData
from hpcflow.sdk.typing import PathLike, hydrate, ParamSource
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.utils import search_dir_files_by_regex
from hpcflow.sdk.core.zarr_io import zarr_decode
from hpcflow.sdk.core.parameters import _process_demo_data_strings

if TYPE_CHECKING:
    import os
    from collections.abc import Mapping
    from typing import Any, ClassVar
    from typing_extensions import Self
    from .actions import Action, ActionRule
    from .environment import Environment
    from .object_list import CommandFilesList
    from .parameters import Parameter
    from .task import ElementSet
    from .workflow import Workflow


class FileNamePart(Protocol):
    """
    A filename or piece of filename that can be expanded.
    """

    def value(self, directory: str | os.PathLike = ".") -> str | list[str]:
        """
        Get the part of the file, possibly with directory specified.
        Implementations of this may ignore the directory.
        If a pattern, the expanded value may be a list of strings.
        """


@dataclass(init=False)
@hydrate
class FileSpec(JSONLike):
    """
    A specification of a file handled by a workflow.
    """

    _validation_schema: ClassVar[str] = "files_spec_schema.yaml"
    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(name="name", class_name="FileNameSpec"),
    )

    #: Label for this file specification.
    label: Final[str]
    #: The name of the file.
    name: Final[FileNameSpec]
    #: Documentation for the file specification.
    doc: Final[str]
    _hash_value: str | None = field(default=None, repr=False)

    def __init__(
        self,
        label: str,
        name: str | FileNameSpec,
        doc: str = "",
        _hash_value: str | None = None,
    ) -> None:
        self.label = label
        self.name = self._app.FileNameSpec(name) if isinstance(name, str) else name
        self.doc = doc
        self._hash_value = _hash_value
        self.__hash = hash((label, self.name))

    def value(self, directory: str | os.PathLike = ".") -> str:
        """
        The path to a file, optionally resolved with respect to a particular directory.
        """
        return cast("str", self.name.value(directory))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.label == other.label and self.name == other.name

    def __hash__(self) -> int:
        return self.__hash

    def _postprocess_to_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        d.pop("_FileSpec__hash")
        return d

    @property
    def stem(self) -> FileNameStem:
        """
        The stem of the file name.
        """
        return self.name.stem

    @property
    def ext(self) -> FileNameExt:
        """
        The extension of the file name.
        """
        return self.name.ext

    @property
    def documentation(self) -> str:
        """
        Documentation for rendering via Jinja.
        """
        if self.doc:
            import markupsafe

            return markupsafe.Markup(self.doc)
        return repr(self)


@hydrate
class FileNameSpec(JSONLike):
    """
    The name of a file handled by a workflow, or a pattern that matches multiple files.

    Parameters
    ----------
    name: str
        The name or pattern.
    args: list
        Positional arguments to use when formatting the name.
        Can be omitted if the name does not contain a Python formatting pattern.
    is_regex: bool
        If true, the name is used as a regex to search for actual files.
    """

    def __init__(
        self,
        name: str,
        args: list[FileNamePart] | None = None,
        is_regex: bool = False,
    ) -> None:
        #: The name or pattern.
        self.name: Final[str] = name
        #: Positional arguments to use when formatting the name.
        self.args: Final[tuple[FileNamePart, ...]] = tuple(args or [])
        #: Whether the name is used as a regex to search for actual files.
        self.is_regex: Final[bool] = is_regex
        self.__hash = hash((name, self.args, is_regex))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            self.name == other.name
            and self.args == other.args
            and self.is_regex == other.is_regex
        )

    def __hash__(self) -> int:
        return self.__hash

    def _postprocess_to_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        d.pop("_FileNameSpec__hash")
        return d

    @property
    def stem(self) -> FileNameStem:
        """
        The stem of the name or pattern.
        """
        return self._app.FileNameStem(self)

    @property
    def ext(self) -> FileNameExt:
        """
        The extension of the name or pattern.
        """
        return self._app.FileNameExt(self)

    def value(self, directory: str | os.PathLike = ".") -> list[str] | str:
        """
        Get the template-resolved name of the file
        (or files matched if the name is a regex pattern).

        Parameters
        ----------
        directory: PathLike
            Where to resolve values with respect to.
        """
        format_args = [arg.value(Path(directory)) for arg in self.args]
        value = self.name.format(*format_args)
        if self.is_regex:
            return search_dir_files_by_regex(value, directory=directory)
        return value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"


@dataclass
class FileNameStem(JSONLike):
    """
    The stem of a file name.
    """

    #: The file specification this is derived from.
    file_name: FileNameSpec

    def value(self, directory: str | os.PathLike = ".") -> str:
        """
        Get the stem, possibly with directory specified.
        """
        d = self.file_name.value(directory)
        if self.file_name.is_regex:
            raise ValueError("cannot get the stem of a regex match")
        assert not isinstance(d, list)
        return Path(d).stem


@dataclass
class FileNameExt(JSONLike):
    """
    The extension of a file name.
    """

    #: The file specification this is derived from.
    file_name: FileNameSpec

    def value(self, directory: str | os.PathLike = ".") -> str:
        """
        Get the extension.
        """
        d = self.file_name.value(directory)
        if self.file_name.is_regex:
            raise ValueError("cannot get the extension of a regex match")
        assert not isinstance(d, list)
        return Path(d).suffix


@dataclass
@hydrate
class InputFileGenerator(JSONLike):
    """
    Represents a script that is run to generate input files for an action.

    Parameters
    ----------
    input_file:
        The file to generate.
    inputs: list[~hpcflow.app.Parameter]
        The input parameters to the generator.
    script:
        The script that generates the input.
    environment:
        The environment in which to run the generator.
    script_pass_env_spec:
        Whether to pass in the environment.
    abortable:
        Whether the generator can be stopped early.
        Quick-running scripts tend to not need this.
    rules: list[~hpcflow.app.ActionRule]
        User-specified rules for whether to run the generator.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="input_file",
            class_name="FileSpec",
            shared_data_primary_key="label",
            shared_data_name="command_files",
        ),
        ChildObjectSpec(
            name="inputs",
            class_name="Parameter",
            is_multiple=True,
            json_like_name="from_inputs",
            shared_data_primary_key="typ",
            shared_data_name="parameters",
        ),
        ChildObjectSpec(
            name="rules",
            class_name="ActionRule",
            is_multiple=True,
            parent_ref="input_file_generator",
        ),
    )

    #: The file to generate.
    input_file: FileSpec
    #: The input parameters to the generator.
    inputs: list[Parameter]
    #: The script that generates the inputs.
    script: str | None = None
    #: Information about data input to the script.
    script_data_in: str | Mapping[str, str | ActionData] | None = None
    #: Information about data output from the script.
    script_data_out: str | Mapping[str, str | ActionData] | None = None
    #: The environment in which to run the generator.
    environment: Environment | None = None
    #: Whether to pass in the environment.
    script_pass_env_spec: bool = False
    #: The builtin path to a template that generates the input file.
    jinja_template: str | None = None
    #: The external path to a template that generates the input file.
    jinja_template_path: str | None = None
    #: Whether the generator can be stopped early.
    #: Quick-running scripts tend to not need this.
    abortable: bool = False
    #: User-specified rules for whether to run the generator.
    rules: list[ActionRule] = field(default_factory=list)
    #: Whether the generator requires a working directory.
    requires_dir: bool = True

    def get_action_rules(self) -> list[ActionRule]:
        """
        Get the rules that allow testing if this input file generator must be run or
        not for a given element.
        """
        return [
            self._app.ActionRule.check_missing(f"input_files.{self.input_file.label}")
        ] + self.rules


@dataclass
@hydrate
class OutputFileParser(JSONLike):
    """
    Represents a script that is run to parse output files from an action and create outputs.

    Parameters
    ----------
    output_files: list[FileSpec]
        The output files that this parser will parse.
    output: ~hpcflow.app.Parameter
        The singular output parsed by this parser. Not to be confused with `outputs` (plural).
    script: str
        The name of the file containing the output file parser source.
    environment: ~hpcflow.app.Environment
        The environment to use to run the parser.
    inputs: list[str]
        The other inputs to the parser.
    outputs: list[str]
        Optional multiple outputs from the upstream actions of the schema that are
        required to parametrise this parser.
    options: dict
        Miscellaneous options.
    script_pass_env_spec: bool
        Whether to pass the environment specifier to the script.
    abortable: bool
        Whether this script can be aborted.
    save_files: list[str]
        The files that should be saved to the persistent store for the workflow.
    clean_up: list[str]
        The files that should be immediately removed.
    rules: list[~hpcflow.app.ActionRule]
        Rules for whether to enable this parser.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="output",
            class_name="Parameter",
            shared_data_name="parameters",
            shared_data_primary_key="typ",
        ),
        ChildObjectSpec(
            name="output_files",
            json_like_name="from_files",
            class_name="FileSpec",
            is_multiple=True,
            shared_data_primary_key="label",
            shared_data_name="command_files",
        ),
        ChildObjectSpec(
            name="save_files",
            class_name="FileSpec",
            is_multiple=True,
            shared_data_primary_key="label",
            shared_data_name="command_files",
        ),
        ChildObjectSpec(
            name="clean_up",
            class_name="FileSpec",
            is_multiple=True,
            shared_data_primary_key="label",
            shared_data_name="command_files",
        ),
        ChildObjectSpec(
            name="rules",
            class_name="ActionRule",
            is_multiple=True,
            parent_ref="output_file_parser",
        ),
    )

    #: The output files that this parser will parse.
    output_files: list[FileSpec]
    #: The singular output parsed by this parser.
    #: Not to be confused with :py:attr:`outputs` (plural).
    output: Parameter | None = None
    #: The name of the file containing the output file parser source.
    script: str | None = None
    #: The environment to use to run the parser.
    environment: Environment | None = None
    #: The other inputs to the parser.
    inputs: list[str] | None = None
    #: Optional multiple outputs from the upstream actions of the schema that are
    #: required to parametrise this parser.
    #: Not to be confused with :py:attr:`output` (singular).
    outputs: list[str] | None = None
    #: Miscellaneous options.
    options: dict[str, Any] | None = None
    #: Whether to pass the environment specifier to the script.
    script_pass_env_spec: bool = False
    #: Whether this script can be aborted.
    abortable: bool = False
    #: The files that should be saved to the persistent store for the workflow.
    save_files: InitVar[list[FileSpec] | bool] = True
    _save_files: list[FileSpec] = field(init=False)
    #: The files that should be immediately removed.
    clean_up: list[str] = field(default_factory=list)
    #: Rules for whether to enable this parser.
    rules: list[ActionRule] = field(default_factory=list)
    #: Whether the parser requires a working directory.
    requires_dir: bool = True

    def __post_init__(self, save_files: list[FileSpec] | bool) -> None:
        if not save_files:
            # save no files
            self._save_files = []
        elif save_files is True:
            # save all output files
            self._save_files = [out_f for out_f in self.output_files]
        else:
            self._save_files = save_files

    @override
    def _postprocess_to_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        d = super()._postprocess_to_dict(d)
        if "_save_files" in d:
            d["save_files"] = d.pop("_save_files")
        return d

    @classmethod
    def from_json_like(  # type: ignore[override]
        cls, json_like: dict[str, Any], shared_data: Mapping | None = None
    ) -> Self:
        if "save_files" in json_like:
            if not json_like["save_files"]:
                json_like["save_files"] = []
            elif json_like["save_files"] is True:
                json_like["save_files"] = [i for i in json_like["output_files"]]
        return super().from_json_like(json_like, shared_data)

    def get_action_rules(self) -> list[ActionRule]:
        """Get the rules that allow testing if this output file parser must be run or not
        for a given element."""
        return [
            self._app.ActionRule.check_missing(f"output_files.{out_f.label}")
            for out_f in self.output_files
        ] + self.rules


@hydrate
class _FileContentsSpecifier(JSONLike):
    """Class to represent the contents of a file, either via a file-system path or
    directly."""

    #: What file is this? Only if known.
    file: FileSpec

    def __init__(
        self,
        path: Path | str | None = None,
        contents: str | None = None,
        extension: str = "",
        store_contents: bool = True,
    ) -> None:
        if path is not None and contents is not None:
            raise ValueError("Specify exactly one of `path` and `contents`.")

        if contents is not None and not store_contents:
            raise ValueError(
                "`store_contents` cannot be set to False if `contents` was specified."
            )

        self._path = _process_demo_data_strings(self._app, path)
        self._contents = contents
        self._extension = extension
        self._store_contents = store_contents

        # assigned by `make_persistent`
        self._workflow: Workflow | None = None
        self._value_group_idx: int | None = None

        # assigned by parent `ElementSet`
        self._element_set: ElementSet | None = None

    def __deepcopy__(self, memo: dict | None) -> Self:
        kwargs = self.to_dict()
        value_group_idx = kwargs.pop("value_group_idx")
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._value_group_idx = value_group_idx
        obj._workflow = self._workflow
        obj._element_set = self._element_set
        return obj

    @property
    def normalised_path(self) -> str:
        """
        Full workflow value path to the file.

        Note
        ----
        This is not the same as the path in the filesystem, but is closely
        related.
        """
        return str(self._path) if self._path else "."

    @override
    def _postprocess_to_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        out = super()._postprocess_to_dict(d)
        if "_workflow" in out:
            del out["_workflow"]
        return {k.lstrip("_"): v for k, v in out.items()}

    @classmethod
    def _json_like_constructor(cls, json_like: dict[str, Any]) -> Self:
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""

        _value_group_idx = json_like.pop("value_group_idx", None)
        obj = cls(**json_like)
        obj._value_group_idx = _value_group_idx

        return obj

    def _get_members(self, ensure_contents: bool = False) -> dict[str, Any]:
        out = self.to_dict()
        del out["value_group_idx"]

        if ensure_contents and self._store_contents and self._contents is None:
            out["contents"] = self.read_contents()

        return out

    def make_persistent(
        self,
        workflow: Workflow,
        source: ParamSource,
    ) -> tuple[str, list[int], bool]:
        """Save to a persistent workflow.

        Returns
        -------
        String is the data path for this task input and integer list
        contains the indices of the parameter data Zarr groups where the data is
        stored.

        """

        if self._value_group_idx is not None:
            data_ref = self._value_group_idx
            is_new = False
            if not workflow.check_parameters_exist(data_ref):
                raise RuntimeError(
                    f"{self.__class__.__name__} has a data reference "
                    f"({data_ref}), but does not exist in the workflow."
                )
            # TODO: log if already persistent.
        else:
            data_ref = workflow._add_file(
                store_contents=self.store_contents,
                is_input=True,
                source=source,
                path=self.path,
                contents=self.contents,
                filename=self.file.name.name,
            )
            # data_ref = workflow._add_parameter_data(
            #     data=self._get_members(ensure_contents=True, use_file_label=True),
            #     source=source,
            # )
            is_new = True
            self._value_group_idx = data_ref
            self._workflow = workflow
            self._path = None
            self._contents = None
            self._extension = ""
            self._store_contents = True

        return (self.normalised_path, [data_ref], is_new)

    @overload
    def _get_value(self, value_name: None = None) -> dict[str, Any]: ...

    @overload
    def _get_value(self, value_name: str) -> Any: ...

    def _get_value(self, value_name: str | None = None) -> Any:
        # TODO: fix
        assert self._value_group_idx is None
        if self._value_group_idx is not None:
            from ..persistence.zarr import ZarrPersistentStore

            assert isinstance(self.workflow._store, ZarrPersistentStore)
            # FIXME: Next two lines are both thoroughly broken, but at least resolve to something
            grp = self.workflow._store._get_parameter_group(self._value_group_idx)
            val = zarr_decode(grp)
        else:
            val = self._get_members(ensure_contents=(value_name == "contents"))
        if value_name:
            return val.get(value_name)

        return val

    def read_contents(self) -> str:
        """
        Get the actual contents of the file.
        """
        with self.__path.open("r") as fh:
            return fh.read()

    @property
    def __path(self) -> Path:
        path = self._get_value("path")
        assert path is not None
        return Path(path)

    @property
    def path(self) -> Path | None:
        """
        The path to the file.
        """
        path = self._get_value("path")
        return Path(path) if path else None

    @property
    def store_contents(self) -> Any:
        """
        Whether the file's contents are stored in the workflow's persistent store.
        """
        return self._get_value("store_contents")

    @property
    def contents(self) -> str:
        """
        The contents of the file.
        """
        if self.store_contents:
            return self._get_value("contents")
        else:
            return self.read_contents()

    @property
    def extension(self) -> str:
        """
        The extension of the file.
        """
        return self._get_value("extension")

    @property
    def workflow(self) -> Workflow:
        """
        The owning workflow.
        """
        if self._workflow:
            return self._workflow
        elif self._element_set:
            w_tmpl = self._element_set.task_template.workflow_template
            if w_tmpl and w_tmpl.workflow:
                return w_tmpl.workflow
        raise NotImplementedError


@hydrate
class InputFile(_FileContentsSpecifier):
    """
    An input file.

    Parameters
    ----------
    file:
        What file is this?
    path: Path
        Where is the (original) file?
    contents: str
        What is the contents of the file (if already known)?
    extension: str
        What is the extension of the file?
    store_contents: bool
        Are the file's contents to be cached in the workflow persistent store?
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="file",
            class_name="FileSpec",
            shared_data_name="command_files",
            shared_data_primary_key="label",
        ),
    )

    def __init__(
        self,
        file: FileSpec | str,
        path: Path | str | None = None,
        contents: str | None = None,
        extension: str = "",
        store_contents: bool = True,
    ) -> None:
        if not isinstance(file, FileSpec):
            files: CommandFilesList = self._app.command_files
            self.file = files.get(file)
        else:
            self.file = file

        super().__init__(path, contents, extension, store_contents)

    def _get_members(
        self, ensure_contents: bool = False, use_file_label: bool = False
    ) -> dict[str, Any]:
        out = super()._get_members(ensure_contents)
        if use_file_label:
            out["file"] = self.file.label
        return out

    def __repr__(self) -> str:
        val_grp_idx = ""
        if self._value_group_idx is not None:
            val_grp_idx = f", value_group_idx={self._value_group_idx}"

        path_str = ""
        if self.path is not None:
            path_str = f", path={self.path!r}"

        return (
            f"{self.__class__.__name__}("
            f"file={self.file.label!r}"
            f"{path_str}"
            f"{val_grp_idx}"
            f")"
        )

    @property
    def normalised_files_path(self) -> str:
        """
        Standard name for the file within the workflow.
        """
        return self.file.label

    @property
    def normalised_path(self) -> str:
        return f"input_files.{self.normalised_files_path}"


@hydrate
class InputFileGeneratorSource(_FileContentsSpecifier):
    """
    The source of code for use in an input file generator.

    Parameters
    ----------
    generator:
        How to generate the file.
    path:
        Path to the file to generate.
    contents:
        Contents of the file. Only used when recreating this object.
    extension:
        File name extension.
    """

    def __init__(
        self,
        generator: InputFileGenerator,
        path: Path | str | None = None,
        contents: str | None = None,
        extension: str = "",
    ):
        #: How to generate the file.
        self.generator = generator
        super().__init__(path, contents, extension)


@hydrate
class OutputFileParserSource(_FileContentsSpecifier):
    """
    The source of code for use in an output file parser.

    Parameters
    ----------
    parser:
        How to parse the file.
    path: Path
        Path to the file to parse.
    contents:
        Contents of the file. Only used when recreating this object.
    extension:
        File name extension.
    """

    def __init__(
        self,
        parser: OutputFileParser,
        path: Path | str | None = None,
        contents: str | None = None,
        extension: str = "",
    ):
        #: How to parse the file.
        self.parser = parser
        super().__init__(path, contents, extension)



import numpy as np

from hpcflow.sdk.utils.arrays import get_1D_idx, get_2D_idx, split_arr


def test_get_2D_idx():
    assert get_2D_idx(0, num_cols=10) == (0, 0)
    assert get_2D_idx(4, num_cols=10) == (0, 4)
    assert get_2D_idx(9, num_cols=10) == (0, 9)
    assert get_2D_idx(10, num_cols=10) == (1, 0)
    assert get_2D_idx(13, num_cols=10) == (1, 3)
    assert get_2D_idx(20, num_cols=10) == (2, 0)
    arr_r, arr_c = get_2D_idx(np.array([0, 4, 9, 10, 13, 20]), num_cols=10)
    assert np.array_equal(arr_r, np.array([0, 0, 0, 1, 1, 2]))
    assert np.array_equal(arr_c, np.array([0, 4, 9, 0, 3, 0]))


def test_get_1D_idx():
    assert get_1D_idx(*(0, 0), num_cols=10) == 0
    assert get_1D_idx(*(0, 4), num_cols=10) == 4
    assert get_1D_idx(*(0, 9), num_cols=10) == 9
    assert get_1D_idx(*(1, 0), num_cols=10) == 10
    assert get_1D_idx(*(1, 3), num_cols=10) == 13
    assert get_1D_idx(*(2, 0), num_cols=10) == 20

    assert np.array_equal(
        get_1D_idx(
            np.array([0, 0, 0, 1, 1, 2]), np.array([0, 4, 9, 0, 3, 0]), num_cols=10
        ),
        np.array([0, 4, 9, 10, 13, 20]),
    )


def test_split_arr():
    splt = split_arr(np.array([4, 0, 1, 2, 3, 4, 1, 4, 5, 6]), metadata_size=1)
    assert len(splt) == 2
    assert np.array_equal(splt[0][0], np.array([0]))
    assert np.array_equal(splt[0][1], np.array([1, 2, 3]))
    assert np.array_equal(splt[1][0], np.array([1]))
    assert np.array_equal(splt[1][1], np.array([4, 5, 6]))



import numpy as np
from hpcflow.app import app as hf
from hpcflow.sdk.core.test_utils import make_schemas, make_workflow
from hpcflow.sdk.submission.jobscript import is_jobscript_array, resolve_jobscript_blocks

import pytest


def test_resolve_jobscript_blocks():
    # separate jobscripts due to `is_array`:
    jobscripts = {
        0: {"is_array": True, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": True, "resource_hash": 0, "dependencies": {0: "DEP_DATA"}},
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": True, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": True,
            "blocks": [{"dependencies": {(0, 0): "DEP_DATA"}}],
        },
    ]

    # separate jobscripts due to different `resource_hash`:
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 1, "dependencies": {0: "DEP_DATA"}},
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": False,
            "blocks": [{"dependencies": {(0, 0): "DEP_DATA"}}],
        },
    ]

    # separate jobscripts due to `is_array`:
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": True, "resource_hash": 0, "dependencies": {0: "DEP_DATA"}},
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": True,
            "blocks": [{"dependencies": {(0, 0): "DEP_DATA"}}],
        },
    ]

    # separate jobscripts due to `is_array`:
    jobscripts = {
        0: {"is_array": True, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 0, "dependencies": {0: "DEP_DATA"}},
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": True, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": False,
            "blocks": [{"dependencies": {(0, 0): "DEP_DATA"}}],
        },
    ]

    # combined jobscript due to same resource_hash, not is_array, and dependencies:
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 0, "dependencies": {0: "DEP_DATA"}},
        2: {"is_array": False, "resource_hash": 0, "dependencies": {1: "DEP_DATA"}},
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {
            "resources": None,
            "is_array": False,
            "blocks": [
                {"dependencies": {}},
                {"dependencies": {(0, 0): "DEP_DATA"}},
                {"dependencies": {(0, 1): "DEP_DATA"}},
            ],
        }
    ]

    # combined jobscript due to same resource_hash, not is_array, and dependencies:
    # (checking non-consecutive jobscript index `3` is inconsequential)
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 0, "dependencies": {0: "DEP_DATA"}},
        3: {"is_array": False, "resource_hash": 0, "dependencies": {1: "DEP_DATA"}},
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {
            "resources": None,
            "is_array": False,
            "blocks": [
                {"dependencies": {}},
                {"dependencies": {(0, 0): "DEP_DATA"}},
                {"dependencies": {(0, 1): "DEP_DATA"}},
            ],
        }
    ]

    # jobscript 0 and 1 combined, not 2 due to independence:
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 0, "dependencies": {0: "DEP_DATA"}},
        2: {"is_array": False, "resource_hash": 0, "dependencies": {}},
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {
            "resources": None,
            "is_array": False,
            "blocks": [{"dependencies": {}}, {"dependencies": {(0, 0): "DEP_DATA"}}],
        },
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
    ]

    # separate jobscripts 0,1 due to independence, separate jobscript 2 due to dependence
    # that spans multiple upstream jobscripts that are independent:
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        2: {
            "is_array": False,
            "resource_hash": 0,
            "dependencies": {0: "DEP_DATA", 1: "DEP_DATA"},
        },
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": False,
            "blocks": [{"dependencies": {(0, 0): "DEP_DATA", (1, 0): "DEP_DATA"}}],
        },
    ]

    # combine jobscripts due to dependence
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 0, "dependencies": {0: "DEP_DATA"}},
        2: {
            "is_array": False,
            "resource_hash": 0,
            "dependencies": {0: "DEP_DATA", 1: "DEP_DATA"},
        },
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {
            "resources": None,
            "is_array": False,
            "blocks": [
                {"dependencies": {}},
                {"dependencies": {(0, 0): "DEP_DATA"}},
                {"dependencies": {(0, 0): "DEP_DATA", (0, 1): "DEP_DATA"}},
            ],
        }
    ]

    # separate jobscripts 0,1 due to independence, combined jobscripts 3,4 due to shared
    # dependencies:
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        2: {
            "is_array": False,
            "resource_hash": 0,
            "dependencies": {0: "DEP_DATA", 1: "DEP_DATA"},
        },
        3: {
            "is_array": False,
            "resource_hash": 0,
            "dependencies": {0: "DEP_DATA", 1: "DEP_DATA", 2: "DEP_DATA"},
        },
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": False,
            "blocks": [
                {"dependencies": {(0, 0): "DEP_DATA", (1, 0): "DEP_DATA"}},
                {
                    "dependencies": {
                        (0, 0): "DEP_DATA",
                        (1, 0): "DEP_DATA",
                        (2, 0): "DEP_DATA",
                    }
                },
            ],
        },
    ]

    # seperate jobscripts 0,1,2 due to resource hashes, combined 2,3 due to shared
    # upstream dependencies:
    jobscripts = {
        0: {"is_array": False, "resource_hash": 0, "dependencies": {}},
        1: {"is_array": False, "resource_hash": 1, "dependencies": {0: "DEP_DATA"}},
        2: {"is_array": False, "resource_hash": 0, "dependencies": {1: "DEP_DATA"}},
        3: {
            "is_array": False,
            "resource_hash": 0,
            "dependencies": {0: "DEP_DATA", 2: "DEP_DATA"},
        },
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": False,
            "blocks": [{"dependencies": {(0, 0): "DEP_DATA"}}],
        },
        {
            "resources": None,
            "is_array": False,
            "blocks": [
                {"dependencies": {(1, 0): "DEP_DATA"}},
                {"dependencies": {(0, 0): "DEP_DATA", (2, 0): "DEP_DATA"}},
            ],
        },
    ]

    # test non-consecutive jobscript indices (i.e. 0,1 merged across tasks in previous
    # step); separate jobscripts 0,2,3 due to resource hashes, combined 3,4 due to shared
    # upstream dependencies:
    jobscripts = {
        0: {"resource_hash": 0, "dependencies": {}, "is_array": False},
        2: {
            "resource_hash": 1,
            "dependencies": {0: "DEP_DATA"},
            "is_array": False,
        },
        3: {
            "resource_hash": 0,
            "dependencies": {0: "DEP_DATA", 2: "DEP_DATA"},
            "is_array": False,
        },
        4: {
            "resource_hash": 0,
            "dependencies": {0: "DEP_DATA", 3: "DEP_DATA"},
            "is_array": False,
        },
    }
    assert resolve_jobscript_blocks(jobscripts) == [
        {"resources": None, "is_array": False, "blocks": [{"dependencies": {}}]},
        {
            "resources": None,
            "is_array": False,
            "blocks": [{"dependencies": {(0, 0): "DEP_DATA"}}],
        },
        {
            "resources": None,
            "is_array": False,
            "blocks": [
                {"dependencies": {(0, 0): "DEP_DATA", (1, 0): "DEP_DATA"}},
                {"dependencies": {(0, 0): "DEP_DATA", (2, 0): "DEP_DATA"}},
            ],
        },
    ]


def test_is_job_array_raises_on_bad_scheduler():
    resources = hf.ElementResources(use_job_array=True)
    resources.set_defaults()
    with pytest.raises(ValueError):
        is_jobscript_array(resources=resources, num_elements=2, store=None)


def test_force_array(null_config, tmp_path):
    wk = make_workflow(
        [[{"p1": None}, ("p2",), "t1"]],
        path=tmp_path,
        local_sequences={0: [("inputs.p1", 2, 0)]},
        name="w1",
        overwrite=False,
    )
    sub = wk.add_submission(force_array=True)
    assert len(sub.jobscripts) == 1
    assert sub.jobscripts[0].is_array


def test_merge_jobscript_multi_dependence(null_config, tmp_path):
    s1, s2, s3 = make_schemas(
        ({}, ("p1", "p2"), "t1"),
        (
            {
                "p1": None,
            },
            ("p3",),
            "t2",
        ),
        ({"p1": None, "p3": None}, tuple(), "t3"),
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_merge_js",
        workflow_name="test_merge_js",
        overwrite=True,
        path=tmp_path,
        tasks=[
            hf.Task(schema=s1, repeats=2),
            hf.Task(schema=s2),
            hf.Task(schema=s3),
        ],
    )
    sub = wk.add_submission()
    assert len(sub.jobscripts) == 1
    assert len(sub.jobscripts[0].blocks) == 1


def test_merge_jobscript_multi_dependence_non_array_source(null_config, tmp_path):
    # the second two jobscripts should merge
    s1, s2, s3 = make_schemas(
        ({}, ("p1", "p2"), "t1"),
        (
            {
                "p1": None,
            },
            ("p3",),
            "t2",
        ),
        ({"p1": None, "p3": None}, tuple(), "t3"),
    )
    wk = hf.Workflow.from_template_data(
        template_name="wk_test_merge",
        path=tmp_path,
        tasks=[
            hf.Task(schema=s1),
            hf.Task(schema=s2, repeats=2),
            hf.Task(schema=s3),
        ],
    )
    sub = wk.add_submission(force_array=True)

    assert len(sub.jobscripts) == 2
    assert len(sub.jobscripts[0].blocks) == 1
    assert len(sub.jobscripts[1].blocks) == 1


def test_multi_block_jobscript_multi_dependence(null_config, tmp_path):

    s1, s2, s3, s4 = make_schemas(
        ({"p1": None}, ("p2", "p3"), "t1"),
        ({"p2": None}, ("p4",), "t2"),
        ({"p4": None}, ("p5",), "t3"),
        ({"p3": None, "p5": None}, (), "t4"),
    )
    tasks = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk = hf.Workflow.from_template_data(
        template_name="test_js_blocks",
        workflow_name="test_js_blocks",
        tasks=tasks,
        path=tmp_path,
    )
    sub = wk.add_submission()
    assert len(sub.jobscripts) == 1
    assert len(sub.jobscripts[0].blocks) == 1


def test_multi_block_jobscript_multi_dependence_distinct_resources(null_config, tmp_path):

    s1, s2, s3, s4 = make_schemas(
        ({"p1": None}, ("p2", "p3"), "t1"),
        ({"p2": None}, ("p4",), "t2"),
        ({"p4": None}, ("p5",), "t3"),
        ({"p3": None, "p5": None}, (), "t4"),
    )
    tasks = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk = hf.Workflow.from_template_data(
        template_name="test_js_blocks",
        workflow_name="test_js_blocks",
        tasks=tasks,
        path=tmp_path,
    )
    sub = wk.add_submission()
    assert len(sub.jobscripts) == 3
    assert len(sub.jobscripts[0].blocks) == 1
    assert len(sub.jobscripts[1].blocks) == 1
    assert len(sub.jobscripts[2].blocks) == 2


def test_multi_block_jobscript_multi_dependence_distinct_resources_sequence_and_group(
    null_config, tmp_path
):

    s1, s2, s3 = make_schemas(
        ({"p1": None}, ("p2",), "t1"),
        ({"p2": None}, ("p4",), "t2"),
        ({"p4": None}, ("p5",), "t3"),
    )
    s4 = hf.TaskSchema(
        objective="t4",
        inputs=[hf.SchemaInput("p2", group="g1"), hf.SchemaInput("p5", group="g1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command("echo $((<<sum(parameter:p2)>> + <<sum(parameter:p5)>>))")
                ]
            )
        ],
    )
    tasks = [
        hf.Task(
            schema=s1,
            sequences=[hf.ValueSequence(path="inputs.p1", values=[1, 2])],
            groups=[hf.ElementGroup(name="g1")],
        ),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3, groups=[hf.ElementGroup(name="g1")]),
        hf.Task(schema=s4),
    ]
    wk = hf.Workflow.from_template_data(
        template_name="test_js_blocks",
        workflow_name="test_js_blocks",
        tasks=tasks,
        overwrite=True,
        path=tmp_path,
    )
    sub = wk.add_submission()
    assert len(sub.jobscripts) == 3
    assert len(sub.jobscripts[0].blocks) == 1
    assert len(sub.jobscripts[1].blocks) == 1
    assert len(sub.jobscripts[2].blocks) == 2


def test_combine_scripts_unset_False_jobscript_hash_equivalence(null_config, tmp_path):

    s1 = hf.TaskSchema(
        objective="t1",
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            ),
            hf.Action(commands=[hf.Command(command='echo "hello!"')]),
        ],
    )
    t1 = hf.Task(schema=s1)
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        resources={
            "any": {
                "combine_scripts": False,  # only applies to the Python script action
            },
        },
        template_name="combine_scripts_test",
        path=tmp_path,
    )
    sub = wk.add_submission()

    # test that even though `combine_scripts` is not set on second action (because it is
    # not a Python script action), the resources have an equivalent hash and thus only one
    # jobscript is generated:

    iter_1 = wk.tasks.t1.elements[0].iterations[0]
    act_1 = iter_1.action_runs[0].action
    act_2 = iter_1.action_runs[1].action

    res_1 = iter_1.get_resources_obj(act_1)
    res_2 = iter_1.get_resources_obj(act_2)

    # set to False on first action:
    assert iter_1.get_resources_obj(act_1).combine_scripts == False

    # not set on second action:
    assert iter_1.get_resources_obj(act_2).combine_scripts == None

    # hashes equivalent:
    assert res_1.get_jobscript_hash() == res_2.get_jobscript_hash()

    assert len(sub.jobscripts) == 1


def test_JS_parallelism_default_zarr(null_config, tmp_path):
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 100},
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_JS_parallelism_default_set_zarr",
        path=tmp_path,
        tasks=[t1],
        store="zarr",
    )

    wk.add_submission()  # do not set JS_parallelism

    # zarr supports JS parallelism, so by default should be set to "scheduled":
    assert wk.submissions[0].JS_parallelism == "scheduled"


def test_JS_parallelism_default_json(null_config, tmp_path):
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 100},
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_JS_parallelism_default_set_json",
        path=tmp_path,
        tasks=[t1],
        store="json",
    )

    wk.add_submission()  # do not set JS_parallelism

    # json does not support JS parallelism, so by default should be set to False:
    assert wk.submissions[0].JS_parallelism is False


def test_jobscript_block_run_IDs_equivalence_JSON_Zarr(null_config, tmp_path):
    """The zarr store keeps jobscript-block run IDs in separate arrays, so test
    equivalence."""

    s1, s2, s3, s4 = make_schemas(
        ({"p1": None}, ("p2", "p3"), "t1"),
        ({"p2": None}, ("p4",), "t2"),
        ({"p4": None}, ("p5",), "t3"),
        ({"p3": None, "p5": None}, (), "t4"),
    )
    tasks_zarr = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_zarr = hf.Workflow.from_template_data(
        template_name="test_js_blocks_zarr",
        tasks=tasks_zarr,
        path=tmp_path,
        store="zarr",
    )
    sub_zarr = wk_zarr.add_submission()

    tasks_json = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_json = hf.Workflow.from_template_data(
        template_name="test_js_blocks_json",
        tasks=tasks_json,
        path=tmp_path,
        store="json",
    )
    sub_json = wk_json.add_submission()

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        js_json = sub_json.jobscripts[js_idx]
        assert np.array_equal(js_zarr.all_EAR_IDs, js_json.all_EAR_IDs)

    # reload both workflows from disk, and check again, since above will check data from
    # in-memory modified Submission object
    sub_json = wk_json.reload().submissions[0]
    sub_zarr = wk_zarr.reload().submissions[0]

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        js_json = sub_json.jobscripts[js_idx]
        assert np.array_equal(js_zarr.all_EAR_IDs, js_json.all_EAR_IDs)


def test_jobscript_task_element_maps_equivalence_JSON_Zarr(null_config, tmp_path):
    """The zarr store keeps jobscript-block task-element maps in separate arrays, so test
    equivalence."""

    s1, s2, s3, s4 = make_schemas(
        ({"p1": None}, ("p2", "p3"), "t1"),
        ({"p2": None}, ("p4",), "t2"),
        ({"p4": None}, ("p5",), "t3"),
        ({"p3": None, "p5": None}, (), "t4"),
    )
    tasks_zarr = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_zarr = hf.Workflow.from_template_data(
        template_name="test_js_blocks_zarr",
        tasks=tasks_zarr,
        path=tmp_path,
        store="zarr",
    )
    sub_zarr = wk_zarr.add_submission()

    tasks_json = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_json = hf.Workflow.from_template_data(
        template_name="test_js_blocks_json",
        tasks=tasks_json,
        path=tmp_path,
        store="json",
    )
    sub_json = wk_json.add_submission()

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        assert len(js_zarr.blocks) == len(sub_json.jobscripts[js_idx].blocks)
        for blk_idx, js_blk_zarr in enumerate(js_zarr.blocks):
            js_blk_json = sub_json.jobscripts[js_idx].blocks[blk_idx]
            assert js_blk_zarr.task_elements == js_blk_json.task_elements

    # reload both workflows from disk, and check again, since above will check data from
    # in-memory modified Submission object
    sub_json = wk_json.reload().submissions[0]
    sub_zarr = wk_zarr.reload().submissions[0]

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        assert len(js_zarr.blocks) == len(sub_json.jobscripts[js_idx].blocks)
        for blk_idx, js_blk_zarr in enumerate(js_zarr.blocks):
            js_blk_json = sub_json.jobscripts[js_idx].blocks[blk_idx]
            assert js_blk_zarr.task_elements == js_blk_json.task_elements


def test_jobscript_task_actions_equivalence_JSON_Zarr(null_config, tmp_path):
    """The zarr store keeps jobscript-block task-actions in separate arrays, so test
    equivalence."""

    s1, s2, s3, s4 = make_schemas(
        ({"p1": None}, ("p2", "p3"), "t1"),
        ({"p2": None}, ("p4",), "t2"),
        ({"p4": None}, ("p5",), "t3"),
        ({"p3": None, "p5": None}, (), "t4"),
    )
    tasks_zarr = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_zarr = hf.Workflow.from_template_data(
        template_name="test_js_blocks_zarr",
        tasks=tasks_zarr,
        path=tmp_path,
        store="zarr",
    )
    sub_zarr = wk_zarr.add_submission()

    tasks_json = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_json = hf.Workflow.from_template_data(
        template_name="test_js_blocks_json",
        tasks=tasks_json,
        path=tmp_path,
        store="json",
    )
    sub_json = wk_json.add_submission()

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        assert len(js_zarr.blocks) == len(sub_json.jobscripts[js_idx].blocks)
        for blk_idx, js_blk_zarr in enumerate(js_zarr.blocks):
            js_blk_json = sub_json.jobscripts[js_idx].blocks[blk_idx]
            assert np.array_equal(js_blk_zarr.task_actions, js_blk_json.task_actions)

    # reload both workflows from disk, and check again, since above will check data from
    # in-memory modified Submission object
    sub_json = wk_json.reload().submissions[0]
    sub_zarr = wk_zarr.reload().submissions[0]

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        assert len(js_zarr.blocks) == len(sub_json.jobscripts[js_idx].blocks)
        for blk_idx, js_blk_zarr in enumerate(js_zarr.blocks):
            js_blk_json = sub_json.jobscripts[js_idx].blocks[blk_idx]
            assert np.array_equal(js_blk_zarr.task_actions, js_blk_json.task_actions)


def test_jobscript_dependencies_equivalence_JSON_Zarr(null_config, tmp_path):
    """The zarr store keeps jobscript-block dependencies in separate arrays, so test
    equivalence."""

    s1, s2, s3, s4 = make_schemas(
        ({"p1": None}, ("p2", "p3"), "t1"),
        ({"p2": None}, ("p4",), "t2"),
        ({"p4": None}, ("p5",), "t3"),
        ({"p3": None, "p5": None}, (), "t4"),
    )
    tasks_zarr = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_zarr = hf.Workflow.from_template_data(
        template_name="test_js_blocks_zarr",
        tasks=tasks_zarr,
        path=tmp_path,
        store="zarr",
    )
    sub_zarr = wk_zarr.add_submission()

    tasks_json = [
        hf.Task(schema=s1, inputs={"p1": 101}),
        hf.Task(schema=s2, resources={"any": {"num_cores": 2}}),
        hf.Task(schema=s3),
        hf.Task(schema=s4),
    ]
    wk_json = hf.Workflow.from_template_data(
        template_name="test_js_blocks_json",
        tasks=tasks_json,
        path=tmp_path,
        store="json",
    )
    sub_json = wk_json.add_submission()

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        assert len(js_zarr.blocks) == len(sub_json.jobscripts[js_idx].blocks)
        for blk_idx, js_blk_zarr in enumerate(js_zarr.blocks):
            js_blk_json = sub_json.jobscripts[js_idx].blocks[blk_idx]
            assert js_blk_zarr.dependencies == js_blk_json.dependencies

    # reload both workflows from disk, and check again, since above will check data from
    # in-memory modified Submission object
    sub_json = wk_json.reload().submissions[0]
    sub_zarr = wk_zarr.reload().submissions[0]

    assert len(sub_zarr.jobscripts) == len(sub_json.jobscripts)

    for js_idx, js_zarr in enumerate(sub_zarr.jobscripts):
        assert len(js_zarr.blocks) == len(sub_json.jobscripts[js_idx].blocks)
        for blk_idx, js_blk_zarr in enumerate(js_zarr.blocks):
            js_blk_json = sub_json.jobscripts[js_idx].blocks[blk_idx]
            assert js_blk_zarr.dependencies == js_blk_json.dependencies



from __future__ import annotations
import sys
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
import pytest
import requests

from hpcflow.app import app as hf
from hpcflow.sdk.core.test_utils import P1_parameter_cls as P1
from hpcflow.sdk.core.utils import read_YAML_str

if TYPE_CHECKING:
    from hpcflow.sdk.core.parameters import Parameter


@pytest.fixture
def null_config(tmp_path: Path):
    if not hf.is_config_loaded:
        hf.load_config(config_dir=tmp_path)


@pytest.fixture
def param_p1(null_config) -> Parameter:
    return hf.Parameter("p1")


def test_fix_trailing_path_delimiter(param_p1: Parameter):
    iv1 = hf.InputValue(parameter=param_p1, value=101, path="a.")
    iv2 = hf.InputValue(parameter=param_p1, value=101, path="a")
    assert iv1.path == iv2.path


def test_fix_single_path_delimiter(param_p1: Parameter):
    iv1 = hf.InputValue(parameter=param_p1, value=101, path=".")
    iv2 = hf.InputValue(parameter=param_p1, value=101)
    assert iv1.path == iv2.path


def test_normalised_path_without_path(param_p1: Parameter):
    iv1 = hf.InputValue(parameter=param_p1, value=101)
    assert iv1.normalised_path == "inputs.p1"


def test_normalised_path_with_single_element_path(param_p1: Parameter):
    iv1 = hf.InputValue(parameter=param_p1, value=101, path="a")
    assert iv1.normalised_path == "inputs.p1.a"


def test_normalised_path_with_multi_element_path(param_p1: Parameter):
    iv1 = hf.InputValue(parameter=param_p1, value=101, path="a.b")
    assert iv1.normalised_path == "inputs.p1.a.b"


def test_normalised_path_with_empty_path(param_p1: Parameter):
    iv1 = hf.InputValue(parameter=param_p1, value=101, path="")
    assert iv1.normalised_path == "inputs.p1"


def test_resource_spec_get_param_path(null_config) -> None:
    rs1 = hf.ResourceSpec()
    assert rs1.normalised_path == "resources.any"


def test_resource_spec_get_param_path_scope_any_with_single_kwarg(null_config) -> None:
    rs1 = hf.ResourceSpec(scratch="local")
    assert rs1.normalised_path == "resources.any"


def test_resources_spec_get_param_path_scope_main(null_config) -> None:
    rs1 = hf.ResourceSpec(scope=hf.ActionScope.main())
    assert rs1.normalised_path == "resources.main"


def test_resources_spec_get_param_path_scope_with_kwargs(null_config) -> None:
    rs1 = hf.ResourceSpec(scope=hf.ActionScope.input_file_generator(file="file1"))
    assert rs1.normalised_path == "resources.input_file_generator[file=file1]"


def test_resources_spec_get_param_path_scope_with_no_kwargs(null_config) -> None:
    rs1 = hf.ResourceSpec(scope=hf.ActionScope.input_file_generator())
    assert rs1.normalised_path == "resources.input_file_generator"


def test_input_value_from_json_like_class_method_attribute_is_set(null_config) -> None:
    parameter_typ = "p1c"
    cls_method = "from_data"
    json_like = {"parameter": f"{parameter_typ}::{cls_method}", "value": {"a": 101}}
    inp_val = hf.InputValue.from_json_like(json_like, shared_data=hf.template_components)
    assert inp_val.parameter.typ == parameter_typ
    assert inp_val.value_class_method == cls_method


def test_value_sequence_from_json_like_class_method_attribute_is_set(null_config) -> None:
    parameter_typ = "p1"
    cls_method = "from_data"
    json_like = {
        "path": f"inputs.{parameter_typ}::{cls_method}",
        "values": [101],
        "nesting_order": 0,
    }

    val_seq = hf.ValueSequence.from_json_like(
        json_like, shared_data=hf.template_components
    )
    assert val_seq.value_class_method == cls_method


def test_path_attributes(null_config) -> None:
    inp = hf.InputValue(parameter="p1", value=101, path="a.b")
    assert inp.labelled_type == "p1"
    assert inp.normalised_path == "inputs.p1.a.b"
    assert inp.normalised_inputs_path == "p1.a.b"


def test_path_attributes_with_label_arg(null_config) -> None:
    inp = hf.InputValue(parameter="p1", value=101, path="a.b", label="1")
    assert inp.labelled_type == "p1[1]"
    assert inp.normalised_path == "inputs.p1[1].a.b"
    assert inp.normalised_inputs_path == "p1[1].a.b"


def test_path_attributes_with_label_arg_cast(null_config) -> None:
    inp = hf.InputValue(parameter="p1", value=101, path="a.b", label=1)
    assert inp.labelled_type == "p1[1]"
    assert inp.normalised_path == "inputs.p1[1].a.b"
    assert inp.normalised_inputs_path == "p1[1].a.b"


def test_from_json_like(null_config) -> None:
    inp = hf.InputValue.from_json_like(
        json_like={"parameter": "p1", "value": 101},
        shared_data=hf.template_components,
    )
    assert inp.parameter.typ == hf.Parameter("p1").typ
    assert inp.value == 101
    assert inp.label == ""


def test_from_json_like_with_label(null_config) -> None:
    inp = hf.InputValue.from_json_like(
        json_like={"parameter": "p1[1]", "value": 101},
        shared_data=hf.template_components,
    )
    assert inp.parameter.typ == hf.Parameter("p1").typ
    assert inp.value == 101
    assert inp.label == "1"


def test_value_is_dict_check_success(null_config) -> None:
    # Parameter("p1c") has an associated `ParameterValue` class, so data should be a dict:
    hf.InputValue("p1c", {"a": 101})


def test_value_is_dict_check_raise(null_config) -> None:
    # Parameter("p1c") has an associated `ParameterValue` class so data should be a dict:
    with pytest.raises(ValueError):
        hf.InputValue("p1c", 101)


def test_value_is_dict_check_no_raise_if_sub_parameter(null_config):
    # Parameter("p1c") has an associated `ParameterValue` class, but the specified value
    # is for some sub-data:
    hf.InputValue("p1c", path="a", value=101)


@pytest.mark.xfail(
    condition=sys.platform == "darwin",
    raises=requests.exceptions.HTTPError,
    reason=(
        "GHA MacOS runners use the same IP address, so we get rate limited when "
        "retrieving demo data from GitHub."
    ),
)
def test_demo_data_value(null_config) -> None:
    name = "text_file.txt"
    assert hf.InputValue("p1", value=f"<<demo_data_file:{name}>>").value == str(
        hf.demo_data_cache_dir.joinpath(name)
    )


def test_input_value_from_yaml_and_json_like_various(null_config):

    seed = 9871389
    es_all = [
        dedent(
            """\
    inputs:
      p1: 1
    """
        ),
        dedent(
            f"""\
    inputs:
      p1[A]: 1
    """
        ),
        dedent(
            f"""\
    inputs:
      p1.b: 20
    """
        ),
        dedent(
            f"""\
    inputs:
      p1[A].b: 1
    """
        ),
        dedent(
            f"""\
    inputs:
      p1c::from_data:
        b: 1
        c: 2
    """
        ),
        dedent(
            f"""\
    inputs:
      p1c[A]::from_data:
        b: 1
        c: 2
    """
        ),
        dedent(
            f"""\
    inputs:
      p1::from_normal:
        loc: 1.4
        scale: 0.1
        seed: {seed}
    """
        ),
        dedent(
            f"""\
    inputs:
      p1[A]::from_normal:
        loc: 1.4
        scale: 0.1
        seed: {seed}
    """
        ),
        dedent(
            f"""\
    inputs:
      p1.b::from_normal:
        loc: 1.4
        scale: 0.1
        seed: {seed}
    """
        ),
        dedent(
            f"""\
    inputs:
      p1[A].b::from_normal:
        loc: 1.4
        scale: 0.1
        seed: {seed}
    """
        ),
    ]

    es_JSONs = [read_YAML_str(es_i) for es_i in es_all]
    es = [
        hf.ElementSet.from_json_like(
            es_json_i,
            shared_data=hf.template_components,
        )
        for es_json_i in es_JSONs
    ]
    inps = [es_i.inputs[0] for es_i in es]

    assert (
        inps[0]
        == hf.InputValue.from_json_like(
            {"parameter": "p1", "value": 1},
            shared_data=hf.template_components,
        )
        == hf.InputValue(hf.Parameter("p1"), value=1)
    )

    assert (
        inps[1]
        == hf.InputValue.from_json_like(
            {"parameter": "p1[A]", "value": 1},
            shared_data=hf.template_components,
        )
        == hf.InputValue(hf.Parameter("p1"), label="A", value=1)
    )

    assert (
        inps[2]
        == hf.InputValue.from_json_like(
            {"parameter": "p1.b", "value": 20},
            shared_data=hf.template_components,
        )
        == hf.InputValue(hf.Parameter("p1"), path="b", value=20)
    )

    assert (
        inps[3]
        == hf.InputValue.from_json_like(
            {"parameter": "p1[A].b", "value": 1},
            shared_data=hf.template_components,
        )
        == hf.InputValue(hf.Parameter("p1"), label="A", path="b", value=1)
    )

    assert (
        inps[4]
        == hf.InputValue.from_json_like(
            {"parameter": "p1c::from_data", "value": {"b": 1, "c": 2}},
            shared_data=hf.template_components,
        )
        == hf.InputValue(
            hf.Parameter("p1c"), value={"b": 1, "c": 2}, value_class_method="from_data"
        )
    )

    assert (
        inps[5]
        == hf.InputValue.from_json_like(
            {"parameter": "p1c[A]::from_data", "value": {"b": 1, "c": 2}},
            shared_data=hf.template_components,
        )
        == hf.InputValue(
            hf.Parameter("p1c"),
            label="A",
            value={"b": 1, "c": 2},
            value_class_method="from_data",
        )
    )

    kwargs = {"loc": 1.4, "scale": 0.1, "seed": seed}
    assert (
        inps[6]
        == hf.InputValue.from_json_like(
            {
                "parameter": "p1::from_normal",
                "value": kwargs,
            },
            shared_data=hf.template_components,
        )
        == hf.InputValue.from_normal(
            hf.Parameter("p1"),
            **kwargs,
        )
    )

    assert (
        inps[7]
        == hf.InputValue.from_json_like(
            {
                "parameter": "p1[A]::from_normal",
                "value": kwargs,
            },
            shared_data=hf.template_components,
        )
        == hf.InputValue.from_normal(
            hf.Parameter("p1"),
            label="A",
            **kwargs,
        )
    )

    assert (
        inps[8]
        == hf.InputValue.from_json_like(
            {
                "parameter": "p1.b::from_normal",
                "value": kwargs,
            },
            shared_data=hf.template_components,
        )
        == hf.InputValue.from_normal(
            hf.Parameter("p1"),
            path="b",
            **kwargs,
        )
    )

    assert (
        inps[9]
        == hf.InputValue.from_json_like(
            {
                "parameter": "p1[A].b::from_normal",
                "value": kwargs,
            },
            shared_data=hf.template_components,
        )
        == hf.InputValue.from_normal(
            hf.Parameter("p1"),
            label="A",
            path="b",
            **kwargs,
        )
    )



def main_script_test_direct_in_direct_out_2_fail_allowed(p2):
    # process, accounting for possible unset data:
    p3 = (p2 if p2 is not None else 0) + 100

    # return outputs
    return {"p3": p3}



from contextlib import contextmanager
from pathlib import Path
import sys
from typing import Any


def resolve_path(path):
    """On Windows Python 3.8, 3.9, and 3.10, `Pathlib.resolve` does
    not return an absolute path for non-existant paths, when it should.

    See: https://github.com/python/cpython/issues/82852

    """
    # TODO: this only seems to be used in a test; remove?
    return Path.cwd() / Path(path).resolve()  # cwd is ignored if already absolute


@contextmanager
def override_module_attrs(module_name: str, overrides: dict[str, Any]):
    """Context manager to temporarily override module-level attributes. The module must be
    imported (i.e. within `sys.modules`)."""

    module = sys.modules[module_name]
    original_values = {k: getattr(module, k) for k in overrides}
    try:
        for k, v in overrides.items():
            setattr(module, k, v)
        yield
    finally:
        for k, v in original_values.items():
            setattr(module, k, v)



"""
Configuration system class.
"""

from __future__ import annotations
import contextlib

from copy import deepcopy
import copy
import functools
import json
import logging
import os
import socket
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import cast, overload, TYPE_CHECKING
import fsspec  # type: ignore

from rich.console import Console, Group
from rich.table import Table
from rich.pretty import Pretty
from rich.panel import Panel
from rich import print as rich_print
from fsspec.registry import known_implementations as fsspec_protocols  # type: ignore
from fsspec.implementations.local import LocalFileSystem  # type: ignore
from hpcflow.sdk.core.utils import get_in_container, read_YAML_file, set_in_container

from hpcflow.sdk.core.validation import get_schema, Schema
from hpcflow.sdk.submission.shells import DEFAULT_SHELL_NAMES
from hpcflow.sdk.typing import PathLike

from hpcflow.sdk.config.callbacks import (
    callback_bool,
    callback_lowercase,
    callback_scheduler_set_up,
    callback_supported_schedulers,
    callback_supported_shells,
    callback_update_log_console_level,
    callback_unset_log_console_level,
    callback_vars,
    callback_file_paths,
    exists_in_schedulers,
    set_callback_file_paths,
    check_load_data_files,
    set_scheduler_invocation_match,
    callback_update_log_file_path,
    callback_update_log_file_level,
    callback_unset_log_file_level,
    callback_unset_log_file_path,
    callback_log_file_path,
)
from hpcflow.sdk.config.config_file import ConfigFile
from hpcflow.sdk.config.errors import (
    ConfigChangeInvalidJSONError,
    ConfigChangePopIndexError,
    ConfigChangeTypeInvalidError,
    ConfigChangeValidationError,
    ConfigItemAlreadyUnsetError,
    ConfigItemCallbackError,
    ConfigNonConfigurableError,
    ConfigReadOnlyError,
    ConfigUnknownItemError,
    ConfigUnknownOverrideError,
    ConfigValidationError,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Mapping, Sequence
    from typing import Any, Literal
    from .types import (
        ConfigDescriptor,
        ConfigMetadata,
        DefaultConfiguration,
        SchedulerConfigDescriptor,
        ShellConfigDescriptor,
        GetterCallback,
        SetterCallback,
        UnsetterCallback,
        T,
    )
    from ..app import BaseApp
    from ..core.types import AbstractFileSystem


logger = logging.getLogger(__name__)

_DEFAULT_SHELL = DEFAULT_SHELL_NAMES[os.name]
#: The default configuration descriptor.
DEFAULT_CONFIG: DefaultConfiguration = {
    "invocation": {"environment_setup": None, "match": {}},
    "config": {
        "machine": socket.gethostname(),
        "log_file_path": "logs/<<app_name>>_v<<app_version>>.log",
        "environment_sources": [],
        "task_schema_sources": [],
        "command_file_sources": [],
        "parameter_sources": [],
        "default_scheduler": "direct",
        "default_shell": _DEFAULT_SHELL,
        "schedulers": {"direct": {"defaults": {}}},
        "shells": {_DEFAULT_SHELL: {"defaults": {}}},
        "user_affiliations": [],
    },
}


@dataclass
class ConfigOptions:
    """Application-level options for configuration"""

    #: The default directory.
    default_directory: Path | str
    #: The environment variable containing the directory name.
    directory_env_var: str
    #: The default configuration.
    default_config: DefaultConfiguration = field(
        default_factory=lambda: deepcopy(DEFAULT_CONFIG)
    )
    #: Any extra schemas to apply.
    extra_schemas: Sequence[Schema] = field(default_factory=list)
    #: Default directory of known configurations.
    default_known_configs_dir: str | None = None
    _schemas: Sequence[Schema] = field(init=False)
    _configurable_keys: Sequence[str] = field(init=False)

    def __post_init__(self) -> None:
        self._schemas, self._configurable_keys = self.init_schemas()

    def init_schemas(self) -> tuple[Sequence[Schema], Sequence[str]]:
        """
        Get allowed configurable keys from config schemas.
        """
        cfg_schemas = [get_schema("config_schema.yaml"), *self.extra_schemas]
        cfg_keys: list[str] = []
        for cfg_schema in cfg_schemas:
            for rule in cfg_schema.rules:
                if not rule.path and rule.condition.callable.name == "allowed_keys":
                    cfg_keys.extend(rule.condition.callable.args)

        return (cfg_schemas, cfg_keys)

    def validate(
        self,
        data: T,
        logger: logging.Logger,
        metadata: ConfigMetadata | None = None,
        raise_with_metadata: bool = True,
    ) -> T:
        """Validate configuration items of the loaded invocation."""

        logger.debug("Validating configuration...")
        validated_data = data

        for cfg_schema in self._schemas:
            cfg_validated = cfg_schema.validate(validated_data)
            if not cfg_validated.is_valid:
                if not raise_with_metadata:
                    metadata = None
                raise ConfigValidationError(
                    message=cfg_validated.get_failures_string(),
                    meta_data=metadata,
                )
            validated_data = cfg_validated.cast_data

        logger.debug("Configuration is valid.")
        return validated_data


class Config:
    """
    Application configuration as defined in one or more config files.

    This class supports indexing into the collection of properties via Python dot notation.

    Notes
    -----
    On modifying/setting existing values, modifications are not automatically copied
    to the configuration file; use :meth:`save()` to save to the file. Items in `overrides`
    are not saved into the file.

    `schedulers` is used for specifying the available schedulers on this machine, and the
    default arguments that should be used when initialising the
    :py:class:`Scheduler` object.

    `shells` is used for specifying the default arguments that should be used when
    initialising the :py:class:`Shell` object.

    Parameters
    ----------
    app:
        The main hpcflow application instance.
    config_file:
        The configuration file that contains this config.
    options:
        Configuration options to be applied.
    logger:
        Where to log messages relating to configuration.
    config_key:
        The name of the configuration within the configuration file.
    uid: int
        User ID.
    callbacks: dict
        Overrides for the callback system.
    variables: dict[str, str]
        Variables to substitute when processing the configuration.

    Attributes
    ----------
    user_name: str
        The full name of the user to attribute newly created workflows to.
        Mapped to a field in the configuration file.
    user_orcid: str
        User's ORCID.
        Mapped to a field in the configuration file.
    user_affiliations: list[str]
        User's institutional affiliations.
        Mapped to a field in the configuration file.
    linux_release_file: str
        Where to get the description of the Linux release version data.
        Mapped to a field in the configuration file.
    log_file_level: str
        At what level to do logging to the file.
        Mapped to a field in the configuration file.
    log_console_level: str
        At what level to do logging to the console. Usually coarser than to a file.
        Mapped to a field in the configuration file.
    demo_data_manifest_file: str
        Where the manifest describing the demo data is.
        Mapped to a field in the configuration file.
    """

    def __init__(
        self,
        app: BaseApp,
        config_file: ConfigFile,
        options: ConfigOptions,
        logger: logging.Logger,
        config_key: str | None,
        uid: str | None = None,
        callbacks: dict[str, tuple[GetterCallback, ...]] | None = None,
        variables: dict[str, str] | None = None,
        **overrides,
    ):
        self._app = app
        self._file = config_file
        self._options = options
        self._overrides = overrides
        self._logger = logger
        self._variables = variables or {}

        self._file._configs.append(self)

        self._config_key = self._file.select_invocation(
            configs=self._file.data["configs"],
            run_time_info=self._app.run_time_info.to_dict(),
            path=self._file.path,
            config_key=config_key,
        )

        # Callbacks are run on get:
        self._get_callbacks: dict[str, tuple[GetterCallback, ...]] = {
            "task_schema_sources": (callback_file_paths,),
            "environment_sources": (callback_file_paths,),
            "parameter_sources": (callback_file_paths,),
            "command_file_sources": (callback_file_paths,),
            "log_file_path": (callback_vars, callback_log_file_path),
            "telemetry": (callback_bool,),
            "schedulers": (callback_lowercase, callback_supported_schedulers),
            "shells": (callback_lowercase,),
            "default_scheduler": (callback_lowercase, exists_in_schedulers),
            "default_shell": (callback_lowercase, callback_supported_shells),
            "demo_data_manifest_file": (callback_file_paths,),
            **(callbacks or {}),
        }

        # Set callbacks are run on set:
        self._set_callbacks: dict[str, tuple[SetterCallback, ...]] = {
            "task_schema_sources": (set_callback_file_paths, check_load_data_files),
            "environment_sources": (set_callback_file_paths, check_load_data_files),
            "parameter_sources": (set_callback_file_paths, check_load_data_files),
            "command_file_sources": (set_callback_file_paths, check_load_data_files),
            "default_scheduler": (exists_in_schedulers, set_scheduler_invocation_match),
            "default_shell": (callback_supported_shells,),
            "schedulers": (callback_supported_schedulers, callback_scheduler_set_up),
            "log_file_path": (callback_update_log_file_path,),
            "log_file_level": (callback_update_log_file_level,),
            "log_console_level": (callback_update_log_console_level,),
            "demo_data_manifest_file": (set_callback_file_paths,),
        }

        self._unset_callbacks: dict[str, tuple[UnsetterCallback, ...]] = {
            "log_console_level": (callback_unset_log_console_level,),
            "log_file_level": (callback_unset_log_file_level,),
            "log_file_path": (callback_unset_log_file_path,),
        }

        self._configurable_keys = self._options._configurable_keys
        self._modified_keys: ConfigDescriptor = {}
        self._unset_keys: set[str] = set()

        if any((unknown := name) not in self._configurable_keys for name in overrides):
            raise ConfigUnknownOverrideError(name=unknown)

        host_uid, host_uid_file_path = self._get_user_id()

        metadata: ConfigMetadata = {
            "config_directory": self._file.directory,
            "config_file_name": self._file.path.name,
            "config_file_path": self._file.path,
            "config_file_contents": self._file.contents,
            "config_key": self._config_key,
            "config_schemas": self._options._schemas,
            "invoking_user_id": uid or host_uid,
            "host_user_id": host_uid,
            "host_user_id_file_path": host_uid_file_path,
        }
        self._meta_data = metadata

        # used within context manager `cached_config`:
        self._use_cache = False
        self._config_cache: dict[tuple[str, bool, bool, bool], Any] = {}

        # note: this must go at the end, after all instance attributes have been set!
        self._options.validate(
            data=self.get_all(include_overrides=True),
            logger=self._logger,
            metadata=metadata,
        )

    def __dir__(self) -> Iterator[str]:
        yield from super().__dir__()
        yield from self._all_keys

    @property
    def config_directory(self) -> Path:
        """
        The directory containing the configuration file.
        """
        return self._get("config_directory")

    @property
    def config_file_name(self) -> str:
        """
        The name of the configuration file.
        """
        return self._get("config_file_name")

    @property
    def config_file_path(self) -> Path:
        """
        The full path to the configuration file.
        """
        return self._get("config_file_path")

    @property
    def config_file_contents(self) -> str:
        """
        The cached contents of the configuration file.
        """
        return self._get("config_file_contents")

    @property
    def config_key(self) -> str:
        """
        The primary key to select the configuration within the configuration file.
        """
        return self._get("config_key")

    @property
    def config_schemas(self) -> Sequence[Schema]:
        """
        The schemas that apply to the configuration file.
        """
        return self._get("config_schemas")

    @property
    def invoking_user_id(self) -> str:
        """
        User ID that created the workflow.
        """
        return self._get("invoking_user_id")

    @property
    def host_user_id(self) -> str:
        """
        User ID as understood by the script.
        """
        return self._get("host_user_id")

    @property
    def host_user_id_file_path(self) -> Path:
        """
        Where user ID information is stored.
        """
        return self._get("host_user_id_file_path")

    @property
    def machine(self) -> str:
        """
        Machine to submit to.
        Mapped to a field in the configuration file.
        """
        return self._get("machine")

    @machine.setter
    def machine(self, value: str):
        self._set("machine", value)

    @property
    def log_file_path(self) -> str:
        """
        Where to log to.
        Mapped to a field in the configuration file.
        """
        return self._get("log_file_path")

    @log_file_path.setter
    def log_file_path(self, value: str):
        self._set("log_file_path", value)

    @property
    def environment_sources(self) -> Sequence[Path]:
        """
        Where to get execution environment descriptors.
        Mapped to a field in the configuration file.
        """
        return self._get("environment_sources")

    @environment_sources.setter
    def environment_sources(self, value: Sequence[Path]):
        self._set("environment_sources", value)

    @property
    def task_schema_sources(self) -> Sequence[str]:
        """
        Where to get task schemas.
        Mapped to a field in the configuration file.
        """
        return self._get("task_schema_sources")

    @task_schema_sources.setter
    def task_schema_sources(self, value: Sequence[str]):
        self._set("task_schema_sources", value)

    @property
    def command_file_sources(self) -> Sequence[str]:
        """
        Where to get command files.
        Mapped to a field in the configuration file.
        """
        return self._get("command_file_sources")

    @command_file_sources.setter
    def command_file_sources(self, value: Sequence[str]):
        self._set("command_file_sources", value)

    @property
    def parameter_sources(self) -> Sequence[str]:
        """
        Where to get parameter descriptors.
        Mapped to a field in the configuration file.
        """
        return self._get("parameter_sources")

    @parameter_sources.setter
    def parameter_sources(self, value: Sequence[str]):
        self._set("parameter_sources", value)

    @property
    def default_scheduler(self) -> str:
        """
        The name of the default scheduler.
        Mapped to a field in the configuration file.
        """
        return self._get("default_scheduler")

    @default_scheduler.setter
    def default_scheduler(self, value: str):
        self._set("default_scheduler", value)

    @property
    def default_shell(self) -> str:
        """
        The name of the default shell.
        Mapped to a field in the configuration file.
        """
        return self._get("default_shell")

    @default_shell.setter
    def default_shell(self, value: str):
        self._set("default_shell", value)

    @property
    def schedulers(self) -> Mapping[str, SchedulerConfigDescriptor]:
        """
        Settings for supported scheduler(s).
        Mapped to a field in the configuration file.
        """
        return self._get("schedulers")

    @schedulers.setter
    def schedulers(self, value: Mapping[str, SchedulerConfigDescriptor]):
        self._set("schedulers", value)

    @property
    def shells(self) -> Mapping[str, ShellConfigDescriptor]:
        """
        Settings for supported shell(s).
        Mapped to a field in the configuration file.
        """
        return self._get("shells")

    @shells.setter
    def shells(self, value: Mapping[str, ShellConfigDescriptor]):
        self._set("shells", value)

    @property
    def demo_data_dir(self) -> str | None:
        """
        Location of demo data.
        Mapped to a field in the configuration file.
        """
        return self._get("demo_data_dir")

    @demo_data_dir.setter
    def demo_data_dir(self, value: str | None):
        self._set("demo_data_dir", value)

    def __getattr__(self, name: str):
        if name.startswith("__"):
            raise AttributeError(f"Attribute not known: {name!r}.")
        return self._get(name)

    def __setattr__(self, name: str, value):
        if (
            "_configurable_keys" in self.__dict__
            and name in self.__dict__["_configurable_keys"]
        ):
            self._set(name, value)
        else:
            super().__setattr__(name, value)

    def _disable_callbacks(self, callbacks: Sequence[str]) -> tuple[
        dict[str, tuple[GetterCallback, ...]],
        dict[str, tuple[SetterCallback, ...]],
        dict[str, tuple[UnsetterCallback, ...]],
    ]:
        """
        Disable named get, set, and unset callbacks.

        Returns
        -------
        The original get and set callback dictionaries.
        """
        self._logger.info(f"disabling config callbacks: {callbacks!r}")
        get_callbacks_tmp: dict[str, tuple[GetterCallback, ...]] = {
            k: tuple(cb for cb in v if cb.__name__ not in callbacks)
            for k, v in self._get_callbacks.items()
        }
        set_callbacks_tmp: dict[str, tuple[SetterCallback, ...]] = {
            k: tuple(cb for cb in v if cb.__name__ not in callbacks)
            for k, v in self._set_callbacks.items()
        }
        unset_callbacks_tmp = {
            k: tuple(i for i in v if i.__name__ not in callbacks)
            for k, v in self._unset_callbacks.items()
        }
        get_callbacks = copy.deepcopy(self._get_callbacks)
        set_callbacks = copy.deepcopy(self._set_callbacks)
        unset_callbacks = copy.deepcopy(self._unset_callbacks)
        self._get_callbacks = get_callbacks_tmp
        self._set_callbacks = set_callbacks_tmp
        self._unset_callbacks = unset_callbacks_tmp
        return (get_callbacks, set_callbacks, unset_callbacks)

    @contextlib.contextmanager
    def _without_callbacks(self, *callbacks: str) -> Iterator[None]:
        """Context manager to temporarily exclude named get, set, and unset callbacks."""
        get_cb, set_cb, unset_cb = self._disable_callbacks(callbacks)
        try:
            yield
        finally:
            self._get_callbacks = get_cb
            self._set_callbacks = set_cb
            self._unset_callbacks = unset_cb

    def _validate(self) -> None:
        data = self.get_all(include_overrides=True)
        self._options.validate(
            data=data,
            logger=self._logger,
            metadata=self._meta_data,
            raise_with_metadata=True,
        )

    def _resolve_path(self, path: PathLike) -> PathLike:
        """Resolve a file path, but leave fsspec protocols alone."""
        if path is None:
            return None
        if any(str(path).startswith(i + ":") for i in fsspec_protocols):
            self._logger.debug(
                f"Not resolving path {path!r} because it looks like an `fsspec` URL."
            )
            return path
        real_path = Path(path).expanduser()
        if real_path.is_absolute():
            return real_path
        return self._meta_data["config_directory"].joinpath(real_path)

    def register_config_get_callback(
        self, name: str
    ) -> Callable[[GetterCallback], GetterCallback]:
        """
        Decorator to register a function as a configuration callback for a specified
        configuration item name, to be invoked on `get` of the item.
        """

        def decorator(func: GetterCallback) -> GetterCallback:
            if name in self._get_callbacks:
                self._get_callbacks[name] = self._get_callbacks[name] + (func,)
            else:
                self._get_callbacks[name] = (func,)

            @functools.wraps(func)
            def wrap(config: Config, value: T) -> T:
                return func(config, value)

            return wrap

        return decorator

    def register_config_set_callback(
        self, name: str
    ) -> Callable[[SetterCallback], SetterCallback]:
        """
        Decorator to register a function as a configuration callback for a specified
        configuration item name, to be invoked on `set` of the item.
        """

        def decorator(func: SetterCallback) -> SetterCallback:
            if name in self._set_callbacks:
                self._set_callbacks[name] = self._set_callbacks[name] + (func,)
            else:
                self._set_callbacks[name] = (func,)

            @functools.wraps(func)
            def wrap(config: Config, value: T) -> Any:
                return func(config, value)

            return wrap

        return decorator

    @property
    def _all_keys(self) -> list[str]:
        return [*self._configurable_keys, *self._meta_data]

    @overload
    def get_all(
        self, *, include_overrides: bool = True, as_str: Literal[True]
    ) -> Mapping[str, str]: ...

    @overload
    def get_all(
        self, *, include_overrides: bool = True, as_str: Literal[False] = False
    ) -> Mapping[str, Any]: ...

    def get_all(
        self, *, include_overrides: bool = True, as_str: bool = False
    ) -> Mapping[str, Any]:
        """Get all configurable items."""
        items: dict[str, Any] = {}
        for key in self._configurable_keys:
            if key in self._unset_keys:
                continue
            try:
                if as_str:
                    items[key] = self._get(
                        name=key,
                        include_overrides=include_overrides,
                        raise_on_missing=True,
                        as_str=True,
                    )
                else:
                    items[key] = self._get(
                        name=key,
                        include_overrides=include_overrides,
                        raise_on_missing=True,
                    )
            except ValueError:
                continue
        return items

    def _show(self, config: bool = True, metadata: bool = False):
        group_args: list[Panel] = []
        if metadata:
            tab = Table(show_header=False, box=None)
            tab.add_column()
            tab.add_column()
            for k, v in self._meta_data.items():
                if k == "config_file_contents":
                    continue
                tab.add_row(k, Pretty(v))
            group_args.append(Panel(tab, title="Config metadata"))

        if config:
            tab = Table(show_header=False, box=None)
            tab.add_column()
            tab.add_column()
            for k, v in self.get_all().items():
                tab.add_row(k, Pretty(v))
            group_args.append(Panel(tab, title=f"Config {self._config_key!r}"))

        rich_print(Group(*group_args))

    def _get_callback_value(self, name: str, value):
        if name in self._get_callbacks and value is not None:
            for cb in self._get_callbacks.get(name, ()):
                self._logger.debug(
                    f"Invoking `config.get` callback ({cb.__name__!r}) for item {name!r}={value!r}"
                )
                try:
                    value = cb(self, value)
                except Exception as err:
                    raise ConfigItemCallbackError(name, cb, err) from None
        return value

    @overload
    def _get(
        self,
        name: str,
        *,
        include_overrides=True,
        raise_on_missing=False,
        as_str: Literal[False] = False,
        callback=True,
        default_value=None,
    ) -> Any: ...

    @overload
    def _get(
        self,
        name: str,
        *,
        include_overrides=True,
        raise_on_missing=False,
        as_str: Literal[True],
        callback=True,
        default_value=None,
    ) -> list[str] | str: ...

    def _get(
        self,
        name: str,
        *,
        include_overrides=True,
        raise_on_missing=False,
        as_str=False,
        callback=True,
        default_value=None,
    ):
        """Get a configuration item."""

        if self._use_cache:
            # note: we default_value is not necessarily hashable, so we can't cache on it!
            key = (
                name,
                include_overrides,
                raise_on_missing,
                as_str,
            )
            if key in self._config_cache:
                return self._config_cache[key]

        if name not in self._all_keys:
            raise ConfigUnknownItemError(name=name)

        elif name in self._meta_data:
            val = cast("dict", self._meta_data)[name]

        elif include_overrides and name in self._overrides:
            val = self._overrides[name]

        elif name in self._unset_keys:
            if raise_on_missing:
                raise ValueError("Not set.")
            val = None
            if default_value:
                val = default_value

        elif name in self._modified_keys:
            val = cast("dict", self._modified_keys)[name]

        elif name in self._configurable_keys:
            val = self._file.get_config_item(
                config_key=self._config_key,
                name=name,
                raise_on_missing=raise_on_missing,
                default_value=default_value,
            )

        if callback:
            val = self._get_callback_value(name, val)

        if as_str:
            if isinstance(val, (list, tuple, set)):
                val = [str(i) for i in val]
            else:
                val = str(val)

        if self._use_cache:
            self._config_cache[key] = val

        return val

    def _parse_JSON(self, name: str, value: str) -> Any:
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError as err:
            raise ConfigChangeInvalidJSONError(name=name, json_str=value, err=err)

    @overload
    def _set(
        self, name: str, value: str, *, is_json: Literal[True], callback=True, quiet=False
    ) -> None: ...

    @overload
    def _set(
        self,
        name: str,
        value: Any,
        *,
        is_json: Literal[False] = False,
        callback=True,
        quiet=False,
    ) -> None: ...

    def _set(
        self, name: str, value, *, is_json=False, callback=True, quiet=False
    ) -> None:
        """
        Set a configuration item.
        """
        if self._use_cache:
            raise ConfigReadOnlyError()

        if name not in self._configurable_keys:
            raise ConfigNonConfigurableError(name=name)
        if is_json:
            value = self._parse_JSON(name, cast("str", value))
        current_val = self._get(name)
        callback_val = self._get_callback_value(name, value)
        file_val = self._get_callback_value(
            name, self._file.get_config_item(self._config_key, name)
        )

        if callback_val != current_val:
            was_in_modified = False
            was_in_unset = False
            prev_modified_val = None
            modified_updated = False
            mk = cast("dict", self._modified_keys)

            if name in self._modified_keys:
                was_in_modified = True
                prev_modified_val = mk[name]

            if name in self._unset_keys:
                was_in_unset = True
                self._unset_keys.remove(name)

            if callback_val != file_val:
                mk[name] = value
                modified_updated = True

            try:
                self._validate()

                if callback:
                    for cb in self._set_callbacks.get(name, ()):
                        self._logger.debug(
                            f"Invoking `config.set` callback for item {name!r}: {cb.__name__!r}"
                        )
                        cb(self, callback_val)

            except ConfigValidationError as err:
                # revert:
                if modified_updated:
                    if was_in_modified:
                        mk[name] = prev_modified_val
                    else:
                        del mk[name]
                if was_in_unset:
                    self._unset_keys.add(name)

                raise ConfigChangeValidationError(name, validation_err=err) from None

            self._logger.debug(
                f"Successfully set config item {name!r} to {callback_val!r}."
            )
        elif not quiet:
            print(f"value is already: {callback_val!r}")

    @overload
    def set(
        self,
        path: str,
        value: Any,
        *,
        is_json: Literal[False] = False,
        quiet: bool = False,
    ) -> None: ...

    @overload
    def set(
        self, path: str, value: str, *, is_json: Literal[True], quiet: bool = False
    ) -> None: ...

    def set(
        self, path: str, value: Any, *, is_json: bool = False, quiet: bool = False
    ) -> None:
        """
        Set the value of a configuration item.

        Parameters
        ----------
        path:
            Which configuration item to set.
        value:
            What to set it to.
        """
        self._logger.debug(f"Attempting to set config item {path!r} to {value!r}.")

        if is_json:
            value = self._parse_JSON(path, value)

        name, *path_suffix = path.split(".")
        root = deepcopy(self._get(name, callback=False))
        if path_suffix:
            if root is None:
                root = {}
                self.set(path=name, value={}, quiet=True)
            set_in_container(
                root,
                path=path_suffix,
                value=value,
                ensure_path=True,
                cast_indices=True,
            )
        else:
            root = value
        self._set(name, root, quiet=quiet)

    def unset(self, name: str, callback: bool = True) -> None:
        """
        Unset the value of a configuration item.

        Parameters
        ----------
        name:
            The name of the configuration item.

        Notes
        -----
            Only top level configuration items may be unset.
        """
        if name not in self._configurable_keys:
            raise ConfigNonConfigurableError(name=name)
        if name in self._unset_keys or not self._file.is_item_set(self._config_key, name):
            raise ConfigItemAlreadyUnsetError(name=name)

        self._unset_keys.add(name)
        try:
            self._validate()
            if callback:
                for cb in self._unset_callbacks.get(name, []):
                    self._logger.debug(
                        f"Invoking `config.unset` callback for item {name!r}: "
                        f"{cb.__name__!r}."
                    )
                    cb(self)
        except ConfigValidationError as err:
            self._unset_keys.remove(name)
            raise ConfigChangeValidationError(name, validation_err=err) from None

    @overload
    def get(
        self,
        path: str,
        *,
        callback: bool = True,
        copy: bool = False,
        ret_root_and_parts: Literal[False] = False,
        default: Any | None = None,
    ) -> Any: ...

    @overload
    def get(
        self,
        path: str,
        *,
        callback: bool = True,
        copy: bool = False,
        ret_root_and_parts: Literal[True],
        default: Any | None = None,
    ) -> tuple[Any, Any, list[str]]: ...

    def get(
        self,
        path: str,
        *,
        callback: bool = True,
        copy: bool = False,
        ret_root_and_parts: bool = False,
        default: Any | None = None,
    ) -> Any:
        """
        Get the value of a configuration item.

        Parameters
        ----------
        path:
            The name of or path to the configuration item.
        """
        name, *suffix = parts = path.split(".")
        root = deepcopy(self._get(name, callback=callback))
        try:
            out = get_in_container(root, suffix, cast_indices=True)
        except KeyError:
            out = default
        if copy:
            out = deepcopy(out)
        if not ret_root_and_parts:
            return out
        return out, root, parts

    def append(self, path: str, value, *, is_json: bool = False) -> None:
        """
        Append a value to a list-like configuration item.

        Parameters
        ----------
        path: str
            The name of or path to the configuration item.
        value:
            The value to append.
        """
        if is_json:
            value = self._parse_JSON(path, value)

        existing, root, parts = self.get(
            path,
            ret_root_and_parts=True,
            callback=False,
            default=[],
        )

        try:
            new = existing + [value]
        except TypeError:
            raise ConfigChangeTypeInvalidError(path, typ=type(existing)) from None

        if parts[1:]:
            set_in_container(
                root,
                path=parts[1:],
                value=new,
                ensure_path=True,
                cast_indices=True,
            )
        else:
            root = new
        self._set(parts[0], root)

    def prepend(self, path: str, value, *, is_json: bool = False) -> None:
        """
        Prepend a value to a list-like configuration item.

        Parameters
        ----------
        path: str
            The name of or path to the configuration item.
        value:
            The value to prepend.
        """
        if is_json:
            value = self._parse_JSON(path, value)

        existing, root, parts = self.get(
            path, ret_root_and_parts=True, callback=False, default=[]
        )

        try:
            new = [value] + existing
        except TypeError:
            raise ConfigChangeTypeInvalidError(path, typ=type(existing)) from None

        if parts[1:]:
            set_in_container(
                root,
                path=parts[1:],
                value=new,
                ensure_path=True,
                cast_indices=True,
            )
        else:
            root = new
        self._set(parts[0], root)

    def pop(self, path: str, index) -> None:
        """
        Remove a value from a specified index of a list-like configuration item.

        Parameters
        ----------
        path: str
            The name of or path to the configuration item.
        index: int
            Where to remove the value from. 0 for the first item, -1 for the last.
        """
        existing, root, parts = self.get(
            path,
            ret_root_and_parts=True,
            callback=False,
            default=[],
        )
        new = deepcopy(existing)
        try:
            new.pop(index)
        except AttributeError:
            raise ConfigChangeTypeInvalidError(path, typ=type(existing)) from None
        except IndexError:
            raise ConfigChangePopIndexError(
                path, length=len(existing), index=index
            ) from None

        if parts[1:]:
            set_in_container(
                root,
                path=parts[1:],
                value=new,
                ensure_path=True,
                cast_indices=True,
            )
        else:
            root = new
        self._set(parts[0], root)

    def update(self, path: str, value, *, is_json: bool = False) -> None:
        """
        Update a map-like configuration item.

        Parameters
        ----------
        path: str
            A dot-delimited string of the nested path to update.
        value: dict
            A dictionary to merge in.
        """
        if is_json:
            value = self._parse_JSON(path, value)

        val_mod, root, parts = self.get(
            path,
            copy=True,
            ret_root_and_parts=True,
            callback=False,
            default={},
        )

        try:
            val_mod.update(value)
        except TypeError:
            raise ConfigChangeTypeInvalidError(path, typ=type(val_mod)) from None

        if parts[1:]:
            set_in_container(
                root,
                path=parts[1:],
                value=val_mod,
                ensure_path=True,
                cast_indices=True,
            )
        else:
            root = val_mod
        self._set(parts[0], root)

    def save(self) -> None:
        """Save any modified/unset configuration items into the file."""
        if not self._modified_keys and not self._unset_keys:
            print("No modifications to save!")
        else:
            self._file.save()

    def get_configurable(self) -> Sequence[str]:
        """Get a list of all configurable keys."""
        return self._configurable_keys

    def _get_user_id(self) -> tuple[str, Path]:
        """
        Retrieve (and set if non-existent) a unique user ID that is independent of the
        config directory.
        """

        uid_file_path = self._app.user_data_dir.joinpath("user_id.txt")
        if not uid_file_path.exists():
            uid = str(uuid.uuid4())
            with uid_file_path.open("wt") as fh:
                fh.write(uid)
        else:
            with uid_file_path.open("rt") as fh:
                uid = fh.read().strip()

        return uid, uid_file_path

    def reset(self) -> None:
        """Reset to the default configuration."""
        self._logger.info("Resetting config file to defaults.")
        self._app.reset_config()

    def add_scheduler(self, scheduler: str, **defaults) -> None:
        """
        Add a scheduler.
        """
        if scheduler in self.get("schedulers"):
            print(f"Scheduler {scheduler!r} already exists.")
            return
        self.update(f"schedulers.{scheduler}.defaults", defaults)

    def add_shell(self, shell: str, **defaults) -> None:
        """
        Add a shell.
        """
        if shell in self.get("shells"):
            return
        if shell.lower() == "wsl":
            # check direct_posix scheduler is added:
            self.add_scheduler("direct_posix")
        self.update(f"shells.{shell}.defaults", defaults)

    def add_shell_WSL(self, **defaults) -> None:
        """
        Add shell with WSL prefix.
        """
        if "WSL_executable" not in defaults:
            defaults["WSL_executable"] = "wsl.exe"
        self.add_shell("wsl", **defaults)

    def import_from_file(
        self, file_path: Path | str, *, rename=True, make_new=False
    ) -> None:
        """
        Import config items from a (remote or local) YAML file. Existing config items
        of the same names will be overwritten.

        Parameters
        ----------
        file_path:
            Local or remote path to a config import YAML file which may have top-level
            keys "invocation" and "config".
        rename:
            If True, the current config will be renamed to the stem of the file specified
            in `file_path`. Ignored if `make_new` is True.
        make_new:
            If True, add the config items as a new config, rather than modifying the
            current config. The name of the new config will be the stem of the file
            specified in `file_path`.
        """
        self._logger.debug(f"import from file: {file_path!r}")

        console = Console()
        with console.status(f"Importing config from file {file_path!r}...") as status:
            file_dat: DefaultConfiguration = read_YAML_file(file_path)
            if rename or make_new:
                file_stem = Path(file_path).stem
                name = file_stem
            else:
                name = self._config_key

            obj = self  # `Config` object to update
            if make_new:
                status.update("Adding a new config...")
                # add a new default config:
                self._file.add_default_config(
                    name=file_stem,
                    config_options=self._options,
                )

                # load it:
                new_config_obj = Config(
                    app=self._app,
                    config_file=self._file,
                    options=self._options,
                    config_key=file_stem,
                    logger=self._logger,
                    variables=self._variables,
                )
                obj = new_config_obj

            elif rename:
                if self._config_key != file_stem:
                    self._file.rename_config_key(
                        config_key=self._config_key,
                        new_config_key=file_stem,
                    )

            new_invoc = file_dat.get("invocation")
            new_config = file_dat.get("config", {})

            if new_invoc is not None:
                status.update("Updating invocation details...")
                config_key = file_stem if (make_new or rename) else self._config_key
                obj._file.update_invocation(
                    config_key=config_key,
                    environment_setup=new_invoc.get("environment_setup"),
                    match=new_invoc.get("match", {}),
                )

            # sort in reverse so "schedulers" and "shells" are set before
            # "default_scheduler" and "default_shell" which might reference the former:
            for k, v in sorted(new_config.items(), reverse=True):
                status.update(f"Updating configurable item {k!r}")
                obj.set(k, value=v, quiet=True)

            obj.save()

        print(f"Config {name!r} updated.")

    def init(self, known_name: str, path: str | None = None) -> None:
        """Configure from a known importable config."""
        if not path:
            if not (path := self._options.default_known_configs_dir):
                raise ValueError("Specify an `path` to search for known config files.")
        elif path == ".":
            path = str(Path(path).resolve())

        self._logger.debug(f"init with `path` = {path!r}")

        fs: AbstractFileSystem = fsspec.open(path).fs
        is_local = isinstance(fs, LocalFileSystem)
        local_path = f"{path}/" if is_local else ""
        files = fs.glob(f"{local_path}*.yaml") + fs.glob(f"{local_path}*.yml")
        self._logger.debug(f"All YAML files found in file-system {fs!r}: {files}")

        if not (files := [i for i in files if Path(i).stem.startswith(known_name)]):
            print(f"No configuration-import files found matching name {known_name!r}.")
            return

        print(f"Found configuration-import files: {files!r}")
        for file_i in files:
            path_i = file_i if is_local else f"{path}/{file_i}"
            self.import_from_file(file_path=path_i, make_new=True)

        print("imports complete")
        # if current config is named "default", rename machine to DEFAULT_CONFIG:
        if self._config_key == "default":
            self.set("machine", "DEFAULT_MACHINE")
            self.save()

    def set_github_demo_data_dir(self, sha: str) -> None:
        """
        Set the `demo_data_dir` item, to an fsspec Github URL.

        We use this (via the CLI) when testing the frozen app on Github, because, by
        default, the SHA is set to the current version tag, which might not include recent
        changes to the demo data.
        """
        assert self._app.demo_data_dir is not None
        self.set(
            "demo_data_dir",
            self._app._get_github_url(
                sha=sha, path=self._app.demo_data_dir.replace(".", "/")
            ),
        )

    @contextlib.contextmanager
    def cached_config(self) -> Iterator[None]:
        try:
            self._use_cache = True
            yield
        finally:
            self._use_cache = False
            self._config_cache = {}  # reset the cache

    def _is_set(self, name: str) -> bool:
        """Check if a (non-metadata) config item is set."""
        if name in self._unset_keys:
            return False
        elif name in self._modified_keys:
            return True
        else:
            return self._file.is_item_set(self._config_key, name)

    @contextlib.contextmanager
    def _with_updates(self, updates: dict[str, Any]) -> Iterator[None]:
        # need to run callbacks for unsetting?
        prev_unset = copy.deepcopy(self._unset_keys)
        prev_modified = copy.deepcopy(self._modified_keys)
        to_unset = []
        try:
            for k, v in updates.items():
                if not self._is_set(k):
                    to_unset.append(k)
                self.set(k, v)
            yield
        finally:
            self._unset_keys = prev_unset
            self._modified_keys = prev_modified



import sys

import pytest
from hpcflow.sdk.core.utils import redirect_std_to_file


def test_stdout_redirect(tmp_path):
    file_name = tmp_path / "test.txt"
    expected = "stdout"
    with redirect_std_to_file(file_name, mode="w"):
        print(expected)
    with file_name.open("r") as fp:
        contents = fp.read().strip()
    assert contents == expected


def test_stderr_redirect(tmp_path):
    file_name = tmp_path / "test.txt"
    expected = "stderr"
    with redirect_std_to_file(file_name, mode="w"):
        print(expected, file=sys.stderr)
    with file_name.open("r") as fp:
        contents = fp.read().strip()
    assert contents == expected


def test_exception_exits_with_code(tmp_path):
    file_name = tmp_path / "test.txt"
    with pytest.raises(SystemExit) as exc:
        with redirect_std_to_file(file_name, mode="w"):
            raise ValueError("oh no!")
    assert exc.value.code == 1


def test_exception_prints_to_file(tmp_path):
    file_name = tmp_path / "test.txt"
    with pytest.raises(SystemExit):
        with redirect_std_to_file(file_name, mode="w"):
            raise ValueError("oh no!")
    with file_name.open("r") as fp:
        contents = fp.read().strip()
    assert 'ValueError("oh no!")' in contents


def test_file_not_created(tmp_path):
    file_name = tmp_path / "test.txt"
    assert not file_name.is_file()
    with redirect_std_to_file(file_name, mode="w"):
        pass
    assert not file_name.is_file()
